from flask import request
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal import logger
from portal.rest import rest_helpers
from portal.rest.schemas import SCHEMAS
from portal.service import permissions
from portal.service.auth import find_client
from portal.service.courses import CourseService
from portal.service.errors import ForbiddenError, PortalAPIError
from portal.service.filters import filter_course_dump
from portal.service.general import find_course

courses_namespace = Namespace('courses')
log = logger.get_logger(__name__)


@courses_namespace.route('')
class CourseList(Resource):
    @jwt_required
    # @courses_namespace.response(200, 'Courses list', model=courses_schema)
    @courses_namespace.response(403, 'Not allowed to see courses')
    def get(self):
        # authorization
        permissions.PermissionsService().require.sysadmin()

        courses_list = CourseService().find_all_courses()
        return SCHEMAS.dump('courses', courses_list)

    @jwt_required
    # @courses_namespace.response(200, 'Created course', model=course_schema)
    # @courses_namespace.response(403, 'Not allowed to create course', model=course_schema)
    def post(self):
        permissions.PermissionsService().require.sysadmin()

        data = rest_helpers.parse_request_data(resource='course', action='create')
        new_course = CourseService().create_course(**data)
        return SCHEMAS.dump('course', new_course), 201


@courses_namespace.route('/<string:cid>')
@courses_namespace.param('cid', 'Course id')
@courses_namespace.response(404, 'Course not found')
class CourseResource(Resource):

    @jwt_required
    # @courses_namespace.response(200, 'Course found', model=course_schema)
    @courses_namespace.response(403, 'Not allowed to see course')
    def get(self, cid: str):
        client = find_client()
        course = find_course(cid)
        # authorization
        perm_service = permissions.PermissionsService(course=course)
        if perm_service.check.permissions(['view_course_full']):
            return SCHEMAS.dump('course', course)

        elif perm_service.check.permissions(['view_course_limited']):
            # TODO: Refactor
            dump = SCHEMAS.dump('course', course)
            filtered_course = filter_course_dump(course, dump, client)
            return filtered_course

        raise ForbiddenError(client=client)

    @jwt_required
    @courses_namespace.response(204, 'Course deleted')
    @courses_namespace.response(403, 'Not allowed to delete course')
    def delete(self, cid: str):
        permissions.PermissionsService().require.sysadmin()
        course = find_course(cid)
        CourseService(course=course).delete_course()
        return '', 204

    @jwt_required
    @courses_namespace.response(204, 'Course updated')
    @courses_namespace.response(403, 'Not allowed to update course')
    def put(self, cid: str):
        course = find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.update_course()

        data = rest_helpers.parse_request_data(action='update', resource='course', partial=True)
        CourseService(course=course).update_course(data)
        return '', 204


@courses_namespace.route('/<string:cid>/notes_access_token')
@courses_namespace.param('cid', 'Course id')
@courses_namespace.response(404, 'Course not found')
class CourseNotesToken(Resource):
    @jwt_required
    @courses_namespace.response(
        403, 'Not allowed to manipulate with course\'s notes access token')
    @courses_namespace.response(200, 'Course notes token')
    def get(self, cid):
        course = find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.course_access_token()
        return course.notes_access_token

    @jwt_required
    @courses_namespace.response(204, 'Course\'s notes access token updated')
    @courses_namespace.response(
        403, 'Not allowed to update course\'s notes access token')
    def put(self, cid):
        course = find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.course_access_token()

        json_data = rest_helpers.require_data(action='update_notes_token', resource='course')

        CourseService(course=course).update_notes_token(json_data['token'])
        return '', 204


@courses_namespace.route('/<string:cid>/import')
@courses_namespace.param('cid', 'Course id')
@courses_namespace.response(404, 'Course not found')
class CourseImport(Resource):
    @jwt_required
    # @courses_namespace.response(200, 'Course has been imported', model=course_schema)
    @courses_namespace.response(403, 'Not allowed to import course')
    @courses_namespace.response(400, 'Cannot import course to itself.')
    def put(self, cid: str):
        course = find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.update_course()

        data = rest_helpers.parse_request_data(
            schema=SCHEMAS.course_import, action='import', resource='course'
        )
        source_course = find_course(data['source_course'])
        if source_course == course:
            raise PortalAPIError(
                400, f'[IMPORT] Cannot import course to itself. (id: {cid})')

        config = data['config']

        copied_course = CourseService(course=course).copy_course(course, config)
        return SCHEMAS.dump('course', copied_course)


@courses_namespace.route('/<string:cid>/users')
@courses_namespace.param('cid', 'Course id')
@courses_namespace.response(404, 'Course not found')
class CourseUsers(Resource):
    @jwt_required
    # @courses_namespace.response(200, 'Users in the course', model=users_schema)
    @courses_namespace.response(403, 'Not allowed to see users in the course')
    def get(self, cid):
        course = find_course(cid)
        permissions.PermissionsService(course=course).require.permissions(['view_course_full'])
        group_ids = request.args.getlist('group')
        role_ids = request.args.getlist('role')
        users = CourseService(course=course).get_users_filtered(group_ids, role_ids)
        return SCHEMAS.dump('users', users)
