from flask import request
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal import logger
from portal.database.models import ClientType, ProjectState
from portal.rest import rest_helpers
from portal.rest.schemas import SCHEMAS
from portal.service import auth, general, permissions
from portal.service.errors import ForbiddenError, SubmissionRefusedError
from portal.service.projects import ProjectService
from portal.service.submissions import SubmissionsService
from portal.tools import time

projects_namespace = Namespace('')  # pylint: disable=invalid-name

log = logger.get_logger(__name__)


@projects_namespace.route('/courses/<string:cid>/projects')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.response(404, 'Course not found')
class ProjectsList(Resource):

    @jwt_required
    @projects_namespace.response(404, 'Course not found')
    # @projects_namespace.response(200, 'Projects list', model=projects_schema)
    def get(self, cid: str):
        course = general.find_course(cid)
        projects = ProjectService().list_projects(course)
        return SCHEMAS.dump('projects', projects)

    @jwt_required
    @projects_namespace.response(404, 'Course not found')
    # @projects_namespace.response(201, 'Project created', model=project_schema)
    def post(self, cid: str):
        course = general.find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.update_course()

        data = rest_helpers.parse_request_data(action='create', resource='project')

        new_project = ProjectService().create_project(course, data)
        return SCHEMAS.dump('project', new_project), 201


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectResource(Resource):
    @jwt_required
    # @projects_namespace.response(200, 'Project found', model=project_schema)
    @projects_namespace.response(403, 'Not allowed to get project')
    def get(self, cid: str, pid: str):
        course = general.find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.view_course()

        project = general.find_project(course, pid)
        return SCHEMAS.dump('project', project)

    @jwt_required
    @projects_namespace.response(204, 'Project deleted')
    @projects_namespace.response(403, 'Not allowed to delete project')
    def delete(self, cid: str, pid: str):
        course = general.find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.update_course()

        project = general.find_project(course, pid)
        ProjectService(project).delete_project()
        return '', 204

    @jwt_required
    @projects_namespace.response(204, 'Project updated')
    @projects_namespace.response(403, 'Not allowed to update project')
    def put(self, cid: str, pid: str):
        course = general.find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.write_projects()

        data = rest_helpers.parse_request_data(action='update', resource='project', partial=True)

        project = general.find_project(course, pid)
        ProjectService(project).update_project(data)
        return '', 204


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/config')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectConfigResource(Resource):
    @jwt_required
    # @projects_namespace.response(200, 'Config of the project', model=config_schema)
    # @projects_namespace.response(200, 'Config of the project', model=config_schema_reduced)
    def get(self, cid: str, pid: str):
        course = general.find_course(cid)
        project = general.find_project(course, pid)
        # authorization
        schema = get_config_schema_based_on_permissions(course)
        return schema.dump(project.config)

    @jwt_required
    @projects_namespace.response(204, 'Project config updated')
    def put(self, cid: str, pid: str):
        course = general.find_course(cid)
        # authorization
        permissions.PermissionsService(course=course).require.write_projects()

        data = rest_helpers.parse_request_data(action='update', resource='config')
        project = general.find_project(course, pid)
        ProjectService(project).update_project_config(data)
        return '', 204


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/test-files-refresh')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectTestFilesRefresh(Resource):
    @jwt_required
    @projects_namespace.response(204, 'Project test_files updated')
    def post(self, cid: str, pid: str):
        course = general.find_course(cid)
        project = general.find_project(course, pid)
        # authorization
        permissions.PermissionsService(course=course).require.write_projects()
        ProjectService(project).update_project_test_files()
        return '', 204


@projects_namespace.route(
    '/courses/<string:cid>/projects/<string:pid>/submissions')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectSubmissions(Resource):
    @jwt_required
    # @projects_namespace.response(200, 'Project submissions list', model=submissions_schema)
    def get(self, cid: str, pid: str):
        course = general.find_course(cid)
        # authorization
        perm = ['read_submissions_all']
        permissions.PermissionsService(course=course).require.permissions(perm)

        user_id = request.args.get('user')
        project = general.find_project(course, pid)
        submissions = ProjectService(project).find_project_submissions(user_id)
        return SCHEMAS.dump('submissions', submissions)

    @jwt_required
    # @projects_namespace.response(201, 'Project submission create', model=submission_schema)
    def post(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        user = request.args.get('user')
        user = general.find_user(user) if user is not None else client
        # authorization
        permissions.PermissionsService(course=course).require.create_submission(user)

        # check if a new submission can be created by this user in this project
        project = general.find_project(course, pid)
        _check_submission_create(user, project)

        data = rest_helpers.parse_request_data(
            schema=SCHEMAS.submission_create, action='create', resource='submission'
        )
        log.debug(f"[REST] Create submission: {data}")

        # data for Kontr processing
        service = SubmissionsService()
        new_submission = service.create_submission(
            user=user,
            project=project,
            submission_params=data
        )
        return SCHEMAS.dump('submission', new_submission), 201


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/files')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectTestFiles(Resource):
    def get(self, cid: str, pid: str):
        # Gets test files (feature)
        raise NotImplementedError()


def _check_submission_create(client, project):
    if client.type == ClientType.USER:
        user = general.find_user(client.id)
        # TODO: Might be good to add check whether it is in DEBUG Mode
        if user.is_admin:
            return True
    if project.state(timestamp=time.current_time()) != ProjectState.ACTIVE:
        raise SubmissionRefusedError(f"Project {project.name} not active.")
    if not ProjectService(project).can_create_submission(client):
        raise SubmissionRefusedError(
            f"Submission in project {project.name} already active.")


def get_config_schema_based_on_permissions(course):
    perm_service = permissions.PermissionsService(course=course)
    if perm_service.check.permissions(['view_course_full']):
        return SCHEMAS.config()
    elif perm_service.check.permissions(['view_course_limited']):
        return SCHEMAS.config_reduced()
    else:
        raise ForbiddenError(perm_service.client)
