"""
Groups service
"""

import logging
from typing import List

from portal.database.models import Course, Group, Project, User
from portal.service import filters, general, permissions
from portal.service.errors import ForbiddenError, PortalAPIError
from portal.service.general import find_course, find_group, find_role, get_new_name
from portal.service.users import UserService

log = logging.getLogger(__name__)


def _set_group_data(group: Group, data: dict) -> Group:
    return general.update_entity(group, data, allowed=['name', 'description', 'codename'])


class GroupService:
    def __init__(self, group: Group = None):
        self._group = group

    @property
    def group(self) -> Group:
        return self._group

    def copy_group(self, target: Course, with_users: str):
        """Copies group to a specified course

        Args:
            target(Course): Target course instance
            with_users(str): Whether it should copy with users

        Returns: the new group
        """
        source = self.group
        log.info(
            f"[COPY] Group: {source.id} to course {target.id} with config: {with_users}")
        new_name = get_new_name(source, target)
        new_group = Group(target, codename=new_name)
        new_group.description = source.description
        new_group.name = source.name
        if with_users == "with_users":
            for user in source.users:
                user.groups.append(new_group)
        return new_group

    def delete_group(self):
        """Deletes group
        """
        log.info(f"[DELETE] Group {self.group.id} ({self.group.codename})")
        general.delete_entity(self.group)

    def update_group(self, data: dict) -> Group:
        """Updates group

        Args:
            data(dict): Group data

        Returns(Group): Updated group instance
        """
        log.info(f"[UPDATE] Group {self.group.codename} to {self.group}.")
        return _set_group_data(self.group, data)

    def create_group(self, course: Course, **data) -> Course:
        """Creates a new group

        Args:
            course(Course): Course instance

        Keyword Args:
            name(str): Name of the group
            description(str): Description of the group
            codename(str): Codename of the group

          Returns(Group): Created group instance
          """
        new_group = Group(course=course)
        self._group = new_group
        _set_group_data(new_group, data)
        log.info(
            f"[CREATE] Group for course {course.id} ({course.codename}): {new_group}")
        return new_group

    def update_membership(self, data: dict):
        """Updates the list of members of the group

        Args:
            data(dict): Data
        """
        users_old = set(self.group.users)
        users_add = data.get('add')
        users_remove = data.get('remove')
        users = set()
        if users_add:
            to_add = UserService().find_users(users_add)
            users = users_old.union(to_add)
        if users_remove:
            to_remove = UserService().find_users(users_remove)
            users = users_old.difference(to_remove)
        self.group.users = list(users)
        general.write_entity(self.group)
        log.info(f"[UPDATE] Users of group {self.group.codename} to {self.group.users}.")

    def find_users_by_role(self, course: Course, role_id=None) -> List[User]:
        """Finds users in the group by role

        Args:
            course(Course): Course instance
            role_id(str): Group id

        Returns(list[User]): List of the users
        """
        course = course or self.group.course
        role = find_role(course, role_id) if role_id else None
        return self.group.get_users_based_on_role(role=role)

    def add_user(self, user: User) -> Group:
        """Adds a single user to the group

        Args:
            user(User): User instance

        Returns(Group): updated group instance
        """
        # TODO: Check that user belongs to Course using the Role relationship
        course = self.group.course
        if user not in self.group.users:
            self.group.users.append(user)
            general.write_entity(self.group)
            log.info(
                f"[ADD] Added user {user.username} to group {self.group.codename} "
                f"in course {course.codename}.")
        else:
            log.info(f"[ADD] User {user.username} is already in group {self.group.codename} "
                     f"in course {course.codename}: no change.")
        return self.group

    def remove_user(self, user: User) -> Group:
        """Removes single user from the group

       Args:
           user(User): User instance

       Returns(Group): updated group instance
        """
        course = self.group.course
        try:
            self.group.users.remove(user)
            general.write_entity(self.group)
        except ValueError:
            message = f"[REMOVE] Could not remove user " \
                      f"{user.username} from group {self.group.codename} in course " \
                      f"{course.codename}: group does not contain user."
            raise PortalAPIError(400, message=message)
        log.info(f"[REMOVE] User {user.username} from group {self.group.codename}"
                 f" in course {course.codename}.")
        return self.group

    def import_group(self, data: dict, course: Course = None) -> Group:
        """Imports Group from a course

        Args:
            data(dict): import configuration
            course(Course): Course instance
        """
        course = course or self.group.course
        source_course = find_course(data['source_course'])
        source_group = find_group(source_course, data['source_group'])
        self._group = source_group
        new_group = self.copy_group(course, data['with_users'])
        general.write_entity(new_group)
        log.info(
            f"[IMPORT] Group {source_group.name} from course "
            f"{source_course.codename} to course {course.codename} "
            f"as {new_group.codename}.")
        return new_group

    def list_groups(self, course: Course) -> List[Group]:
        """List all groups
        Args:
            course(Course): Course instance

        Returns(list): List of all groups

        """
        perm_service = permissions.PermissionsService(course=course)
        if perm_service.check.permissions(['view_course_full']):
            return course.groups
        elif perm_service.check.permissions(['view_course_limited']):
            return filters.filter_groups_from_course(course=course, user=perm_service.client)
        raise ForbiddenError(perm_service.client)

    def remove_project(self, project: Project) -> Group:
        """Removes projects from the group
        Args:
            project(Project): Project instance
        Returns:
        """

        course = self.group.course
        try:
            self.group.projects.remove(project)
            general.write_entity(self.group)
        except ValueError:
            message = f"[REMOVE] Could not remove project " \
                      f"{project.codename} from group {self.group.codename} in course " \
                      f"{course.codename}: group does not contain project."
            raise PortalAPIError(400, message=message)
        log.info(f"[REMOVE] Project {project.codename} from group {self.group.codename}"
                 f" in course {course.codename}.")
        return self.group

    def add_project(self, project: Project) -> Group:
        """Adds project to the group
        Args:
=            project(Project): Project instance

        Returns:
        """
        course = self.group.course
        if project not in self.group.projects:
            self.group.projects.append(project)
            general.write_entity(self.group)
            log.info(
                f"[ADD] Added project {project.codename} to group "
                f"{self.group.codename} in course {course.codename}.")
        else:
            log.info(f"[ADD] User {project.codename} is already in group {self.group.codename} "
                     f"in course {course.codename}: no change.")
        return self.group

    def find_projects(self) -> List[Project]:
        """Lists all the projects in the group
        Returns:
        """
        return self.group.projects
